import types, re, copy
from functools import update_wrapper


class Partial:
    """
    A special class wrapper to allow for easy partial function wrappings and calls.
    """

    def __init__(
        self,
        __fn__,
        *__args__,
        **__kwargs__,
    ):
        update_wrapper(self, __fn__)
        self.__fn__ = __fn__
        self.__args__ = __args__
        self.__kwargs__ = __kwargs__
        self.__fnArity__ = self.__getFnArity__()
        self.__arity__ = self.__getArity__(__args__, __kwargs__)

    def __exception__(self, message):
        pre_message = (
            f"({self.__fn__.__module__}.{self.__fn__.__qualname__}_partial): "
        )
        raise Exception(pre_message + message)

    def __call__(self, *args, **kwargs):
        new_args = self.__args__ + args
        new_kwargs = {**self.__kwargs__, **kwargs}
        self.__arity__ = self.__getArity__(new_args, new_kwargs)
        if self.__arity__ < 0:
            self.__exception__("Too many arguments were supplied")
        if self.__arity__ == 0:
            results = self.__fn__(*new_args, **new_kwargs)
            return results
        return Partial(
            self.__fn__,
            *new_args,
            **new_kwargs,
        )

    def __repr__(self):
        return f"<Partial {self.__fn__.__module__}.{self.__fn__.__qualname__} object at {hex(id(self))}>"

    def __getArity__(self, args, kwargs):
        return self.__fnArity__ - (len(args) + len(kwargs))

    def __getFnArity__(self):
        if not isinstance(self.__fn__, (types.MethodType, types.FunctionType)):
            self.__exception__(
                "A non function was passed as a function and does not have any arity. See the stack trace above for more information."
            )
        extra_method_input_count = (
            1 if isinstance(self.__fn__, (types.MethodType)) else 0
        )
        return self.__fn__.__code__.co_argcount - extra_method_input_count


class GenericConstraint:
    def __init__(self, constraints: dict):
        """
        Creates a generic constraint object that can be used to validate a value against a set of constraints.

        Required Arguments:

        - `constraints`:
            - What: A dictionary of constraint names and their associated functions.
            - Type: dict
            - Note: All values in the dictionary must be functions that take a single argument and return a boolean.
            - Note: The dictionary keys will be used to identify the failed constraints in the error messages.
        """
        assert all(
            hasattr(v, "__call__") for v in constraints.values()
        ), "All constraints must be functions."
        self.__constraint_checks__ = constraints

    def __validate__(self, varname, value):
        for check_name, check_func in self.__constraint_checks__.items():
            try:
                if not check_func(value):
                    return f"Constraint `{check_name}` not met with the provided value `{value}`"
            except Exception as e:
                return f"An exception was raised when checking the constraint `{check_name}` with the provided value `{value}`. Error: {e}"
        return True


class Constraint(GenericConstraint):
    def __init__(
        self,
        pattern: [str, None] = None,
        includes: [list, tuple, set, None] = None,
        excludes: [list, tuple, set, None] = None,
        gt: [float, int, None] = None,
        lt: [float, int, None] = None,
        ge: [float, int, None] = None,
        le: [float, int, None] = None,
        eq: [float, int, None] = None,
        ne: [float, int, None] = None,
    ):
        """
        Creates a constraint object that can be used to validate a value against a set of constraints.

        Optional Arguments:

        - `pattern`:
            - What: A regex pattern that the value must match.
            - Type: str or None
            - Default: None
        - `includes`:
            - What: A list of values that the value must be in.
            - Type: list, tuple, set or None
            - Default: None
        - `excludes`:
            - What: A list of values that the value must not be in.
            - Type: list, tuple, set or None
            - Default: None
        - `gt`:
            - What: The value must be greater than this value.
            - Type: float, int or None
            - Default: None
        - `lt`:
            - What: The value must be less than this value.
            - Type: float, int or None
            - Default: None
        - `ge`:
            - What: The value must be greater than or equal to this value.
            - Type: float, int or None
            - Default: None
        - `le`:
            - What: The value must be less than or equal to this value.
            - Type: float, int or None
            - Default: None
        - `eq`:
            - What: The value must be equal to this value.
            - Type: float, int or None
            - Default: None
        - `ne`:
            - What: The value must not be equal to this value.
            - Type: float, int or None
            - Default: None
        """
        assert any(
            v is not None
            for v in [pattern, gt, lt, ge, le, eq, ne, includes, excludes]
        ), "At least one constraint must be provided."
        assert isinstance(
            includes, (list, tuple, set, type(None))
        ), "Includes must be a list, tuple, set or None."
        assert isinstance(
            excludes, (list, tuple, set, type(None))
        ), "Excludes must be a list, tuple, set or None."
        assert isinstance(
            pattern, (str, type(None))
        ), "Pattern must be a string or None."
        assert isinstance(
            gt, (float, int, type(None))
        ), "Greater than constraint must be a float, int or None."
        assert isinstance(
            lt, (float, int, type(None))
        ), "Less than constraint must be a float, int or None."
        assert isinstance(
            ge, (float, int, type(None))
        ), "Greater or equal to constraint must be a float, int or None."
        assert isinstance(
            le, (float, int, type(None))
        ), "Less than or equal to constraint must be a float, int or None."
        assert isinstance(
            eq, (float, int, type(None))
        ), "Equal to constraint must be a float, int or None."
        assert isinstance(
            ne, (float, int, type(None))
        ), "Not equal to constraint must be a float, int or None."
        self.__constraint_checks__ = {}
        if pattern is not None:
            self.__constraint_checks__["be a string"] = lambda x: isinstance(
                x, str
            )
            self.__constraint_checks__["Regex Pattern Match"] = lambda x: bool(
                re.findall(pattern, x)
            )
        if includes is not None:
            self.__constraint_checks__[f"Includes"] = lambda x: x in includes
        if excludes is not None:
            self.__constraint_checks__["Excludes"] = lambda x: x not in excludes
        if gt is not None:
            self.__constraint_checks__[f"Greater Than ({gt})"] = (
                lambda x: x > gt
            )
        if lt is not None:
            self.__constraint_checks__[f"Less Than ({lt})"] = lambda x: x < lt
        if ge is not None:
            self.__constraint_checks__[f"Greater Than Or Equal To ({ge})"] = (
                lambda x: x >= ge
            )
        if le is not None:
            self.__constraint_checks__[f"Less Than Or Equal To ({le})"] = (
                lambda x: x <= le
            )
        if eq is not None:
            self.__constraint_checks__[f"Equal To ({eq})"] = lambda x: x == eq
        if ne is not None:
            self.__constraint_checks__[f"Not Equal To ({ne})"] = (
                lambda x: x != ne
            )


class WithSubclasses:
    def __init__(self, obj, cache=True):
        """
        A special helper class to allow a class type to be passed and also allow all subclasses of that type.

        This allows delayed evaluation of the subclasses until initial call time such that delayed inherited classes are considered.

        Note: You can not validate that a passed object is a WithSubclasses object as this is a special case.

        Required Arguments:

        - `obj`:
            - What: An uninitialized class that should also be considered type correct if a subclass is passed.
            - Type: Any Uninitialized class

        Optional Arguments:

        - `cache`:
            - What: Whether to cache the calcuation of the subclasses. If set to False, this will recalculate the subclasses on each call.
            - Type: bool
            - Default: True
        """
        self.obj = obj
        self.cache = cache

    @classmethod
    def __get_subclasses__(cls, obj):
        """
        A special helper method to get all of the subclasses of a provided class object.

        Requires:

        - `obj`:
            - What: An uninitialized class that should also be considered type correct if a subclass is passed.
            - Type: Any Uninitialized class

        Returns:

        - `out`:
            - What: A list of all of the subclasses (recursively parsed)
            - Type: list of strs


        Notes:

        - From a functional perspective, this recursively gets the subclasses for an uninitialised class (type).
        """
        out = [obj]
        for i in obj.__subclasses__():
            out += cls.__get_subclasses__(i)
        return out

    def get_subclasses(self):
        """
        Returns a list of all of the subclasses of the class that was passed to the WithSubclasses object.

        If already calculated, it will return the cached value.

        Returns:

        - `subclasses`:
            - What: A list of all of the subclasses of the class that was passed to the WithSubclasses object.
            - Type: list of strs
        """
        if not hasattr(self, "subclasses") or self.cache == False:
            self.subclasses = self.__get_subclasses__(self.obj)
        return self.subclasses


def DeepMerge(original: dict, update: dict):
    """
    Merge two dictionaries together, recursively merging any nested dictionaries and extending any nested lists.

    Required Arguments:

    - `original`:
        - What: The original dictionary to merge the update into.
        - Type: dict
    - `update`:
        - What: The dictionary to merge into the original dictionary.
        - Type: dict
    """
    original = copy.deepcopy(original)
    for key, value in update.items():
        if (
            key in original
            and isinstance(original[key], dict)
            and isinstance(value, dict)
        ):
            DeepMerge(original[key], value)
        elif (
            key in original
            and isinstance(original[key], list)
            and isinstance(value, list)
        ):
            original[key].extend(value)
        else:
            original[key] = value
    return original



import type_enforced, sys
from typing import Literal


try:

    @type_enforced.Enforcer
    def my_fn(a: Literal["bar"] | int) -> None:
        pass

    success = True
except:
    success = False

# TODO: In the next major version, add support to enforce combinations of literals as types
#       and not constraints.
# try:
#     my_fn("bar")  # Passes
#     my_fn(1)  # Passes
# except:
#     success = False


if success:
    print(f"test_fn_18.py passed")
elif sys.version_info <= (3, 10, 0):
    print("test_fn_18.py skipped")
else:
    print(f"test_fn_18.py failed")



import type_enforced


@type_enforced.Enforcer
def my_fn(a: dict[dict[int]], b: list[set[str]]) -> None:
    return None


success_1 = True
try:
    my_fn(a={"a": {"a": 1}}, b=[{"a"}])
except:
    success_1 = False

success_2 = False
try:
    my_fn(a={"a": {"a": "2"}}, b=[{"a"}])  # Error (a can only accept int)
except Exception as e:
    if "Type mismatch" in str(e):
        success_2 = True

success_3 = False
try:
    my_fn(a={"a": {"a": 1}}, b=[{"a", 1}])  # Error (b can only accept str)
except Exception as e:
    if "Type mismatch" in str(e):
        success_3 = True

if success_1 and success_2 and success_3:
    print("test_fn_08.py passed")
else:
    print("test_fn_08.py failed")



import type_enforced


@type_enforced.Enforcer
class Foo:
    def bar(self, a: int) -> None:
        pass

    @type_enforced.Enforcer(enabled=False)
    def baz(self, a: int) -> None:
        pass


@type_enforced.Enforcer(enabled=False)
class Boo:
    @type_enforced.Enforcer(enabled=True)
    def bar(self, a: int) -> None:
        pass

    def baz(self, a: int) -> None:
        pass


try:
    foo = Foo()
    foo.bar(a=1)  # => No Exception
    foo.baz(a="a")  # => No Exception

    boo = Boo()
    boo.bar(a=1)  # => No Exception
    boo.baz(a="a")  # => No Exception
    print("test_class_10.py passed")
except:
    print("test_class_10.py failed")



from types import (
    FunctionType,
    MethodType,
    GenericAlias,
    GeneratorType,
    BuiltinFunctionType,
    BuiltinMethodType,
)
from typing import Type, Union, Sized, Literal, Callable
from functools import update_wrapper, wraps
from type_enforced.utils import (
    Partial,
    GenericConstraint,
    DeepMerge,
    WithSubclasses,
)

# Python 3.10+ has a UnionType object that is used to represent Union types
try:
    from types import UnionType
except ImportError:
    # Python < 3.10 does not have UnionType
    # Since UnionType is validated after NoneType, we can use NoneType
    # as a stand in for UnionType to ignore it in older versions of python
    UnionType = type(None)


class FunctionMethodEnforcer:
    def __init__(self, __fn__):
        """
        Initialize a FunctionMethodEnforcer class object as a wrapper for a passed function `__fn__`.

        Requires:

            - `__fn__`:
                - What: The function to enforce
                - Type: function | method | class
        """
        update_wrapper(self, __fn__)
        self.__fn__ = __fn__
        self.__outer_self__ = None
        # Validate that the passed function or method is a method or function
        self.__check_method_function__()
        # Get input defaults for the function or method
        self.__get_defaults__()

    def __get_defaults__(self):
        """
        Get the default values of the passed function or method and store them in `self.__fn_defaults__`.
        """
        self.__fn_defaults__ = {}
        if self.__fn__.__defaults__ is not None:
            # Get the names of all provided default values for args
            default_varnames = list(self.__fn__.__code__.co_varnames)[
                : self.__fn__.__code__.co_argcount
            ][-len(self.__fn__.__defaults__) :]
            # Update the output dictionary with the default values
            self.__fn_defaults__.update(
                dict(zip(default_varnames, self.__fn__.__defaults__))
            )
        if self.__fn__.__kwdefaults__ is not None:
            # Update the output dictionary with the keyword default values
            self.__fn_defaults__.update(self.__fn__.__kwdefaults__)

    def __get_checkable_types__(self):
        """
        Creates two class attributes:

        - `self.__checkable_types__`:
            - What: A dictionary of all annotations as checkable types
            - Type: dict

        - `self.__return_type__`:
            - What: The return type of the function or method
            - Type: dict | None
        """
        if not hasattr(self, "__checkable_types__"):
            self.__checkable_types__ = {
                key: self.__get_checkable_type__(value)
                for key, value in self.__fn__.__annotations__.items()
            }
            self.__return_type__ = self.__checkable_types__.pop("return", None)

    def __get_checkable_type__(self, item_annotation):
        """
        Gets the checkable type as a nested dict for a passed annotation.
        """
        valid_types = {"__extra__": {}}
        if not isinstance(item_annotation, (list, tuple)):
            item_annotation = [item_annotation]
        for valid_type in item_annotation:
            # Special code to replace None with NoneType
            if valid_type is None:
                valid_types[type(None)] = None
                continue
            elif hasattr(valid_type, "__origin__"):
                # Special code for iterable types (e.g. list, tuple, dict, set) including typing iterables
                if valid_type.__origin__ in [list, tuple, dict, set]:
                    valid_types[valid_type.__origin__] = DeepMerge(
                        original=valid_types.get(valid_type.__origin__, {}),
                        update=self.__get_checkable_type__(valid_type.__args__),
                    )
                # Handle any generic aliases
                elif isinstance(valid_type, GenericAlias):
                    valid_types[valid_type.__origin__] = None
                # Handle Union Types (e.g. Union, Optional, ...)
                elif valid_type.__origin__ == Union:
                    valid_types.update(
                        self.__get_checkable_type__(valid_type.__args__)
                    )
                # Handle Literals
                # Note: These will be handled separately in the self.__check_type__
                # as the object is validated and not its type.
                elif valid_type.__origin__ == Literal:
                    valid_types["__extra__"][
                        "__literal__"
                    ] = valid_type.__args__
                # Handle Sized objects
                elif valid_type == Sized:
                    valid_types = {
                        list: None,
                        tuple: None,
                        dict: None,
                        set: None,
                        str: None,
                        bytes: None,
                        bytearray: None,
                        memoryview: None,
                        range: None,
                        # Allow nested types to be passed as well
                        **valid_types,
                    }
                # Handle uninitialized class type objects (e.g. MyCustomClass)
                elif valid_type == Callable:
                    valid_types = {
                        staticmethod: None,
                        classmethod: None,
                        FunctionType: None,
                        BuiltinFunctionType: None,
                        MethodType: None,
                        BuiltinMethodType: None,
                        GeneratorType: None,
                        **valid_types,
                    }
                else:
                    valid_types[valid_type] = None
            # Handle special '|' syntax for Union Types
            elif isinstance(valid_type, UnionType):
                valid_types.update(
                    self.__get_checkable_type__(valid_type.__args__)
                )
            # Handle Constraints
            # Note: These will be handled separately in the self.__check_type__
            # as the object is validated differently than a type.
            elif isinstance(valid_type, GenericConstraint):
                valid_types["__extra__"]["__constraint__"] = valid_type
            elif isinstance(valid_type, WithSubclasses):
                for subclass in valid_type.get_subclasses():
                    valid_types[subclass] = None
            else:
                valid_types[valid_type] = None
        return valid_types

    def __exception__(self, message):
        """
        Usage:

        - Creates a class based exception message

        Requires:

        - `message`:
            - Type: str
            - What: The message to warn users with
        """
        raise TypeError(f"({self.__fn__.__qualname__}): {message}")

    def __get__(self, obj, objtype):
        """
        Overwrite standard __get__ method to return __call__ instead for wrapped class methods.

        Also stores the calling (__get__) `obj` to be passed as an initial argument for `__call__` such that methods can pass `self` correctly.
        """

        @wraps(self.__fn__)
        def __get_fn__(*args, **kwargs):
            return self.__call__(*args, **kwargs)

        self.__outer_self__ = obj
        return __get_fn__

    def __check_method_function__(self):
        """
        Validate that `self.__fn__` is a method or function
        """
        if not isinstance(self.__fn__, (MethodType, FunctionType)):
            raise Exception(
                f"A non function/method was passed to Enforcer. See the stack trace above for more information."
            )

    def __call__(self, *args, **kwargs):
        """
        This method is used to validate the passed inputs and return the output of the wrapped function or method.
        """
        # Special code to pass self as an initial argument
        # for validation purposes in methods
        # See: self.__get__
        if self.__outer_self__ is not None:
            args = (self.__outer_self__, *args)
        # Get a dictionary of all annotations as checkable types
        # Note: This is only done once at first call to avoid redundant calculations
        self.__get_checkable_types__()
        # Create a compreshensive dictionary of assigned variables (order matters)
        assigned_vars = {
            **self.__fn_defaults__,
            **dict(zip(self.__fn__.__code__.co_varnames[: len(args)], args)),
            **kwargs,
        }
        # Validate all listed annotations vs the assigned_vars dictionary
        for key, value in self.__checkable_types__.items():
            self.__check_type__(assigned_vars.get(key), value, key)
        # Execute the function callable
        return_value = self.__fn__(*args, **kwargs)
        # If a return type was passed, validate the returned object
        if self.__return_type__ is not None:
            self.__check_type__(return_value, self.__return_type__, "return")
        return return_value

    def __check_type__(self, obj, acceptable_types, key):
        """
        Raises an exception the type of a passed `obj` (parameter) is not in the list of supplied `acceptable_types` for the argument.
        """
        if isinstance(obj, type):
            passed_type = Type[obj]
        else:
            passed_type = type(obj)
        acceptable_types = dict(acceptable_types)
        extra_types = acceptable_types.pop("__extra__")
        if acceptable_types != {}:
            if passed_type not in acceptable_types:
                # Raise the exception
                self.__exception__(
                    f"Type mismatch for typed variable `{key}`. Expected one of the following `{str(list(acceptable_types.keys()))}` but got `{passed_type}` instead."
                )
        sub_type = acceptable_types.get(passed_type)
        if sub_type is not None:
            if passed_type == dict:
                for sub_key, value in obj.items():
                    self.__check_type__(value, sub_type, f"{key}[{sub_key}]")
            else:
                for sub_key, value in enumerate(obj):
                    self.__check_type__(value, sub_type, f"{key}[{sub_key}]")
        # Special check for Literal types
        literal_options = extra_types.get("__literal__")
        if literal_options is not None:
            if obj not in literal_options:
                self.__exception__(
                    f"Literal validation error for variable `{key}`. Expected one of the following `{str(list(literal_options))}` but got `{obj}` instead."
                )
        # Special check for Constraints
        constraint = extra_types.get("__constraint__")
        if constraint is not None:
            constraint_validation_output = constraint.__validate__(key, obj)
            if constraint_validation_output is not True:
                self.__exception__(
                    f"Constraint validation error for variable `{key}`. {constraint_validation_output}"
                )

    def __repr__(self):
        return f"<type_enforced {self.__fn__.__module__}.{self.__fn__.__qualname__} object at {hex(id(self))}>"


def Enforcer(clsFnMethod, enabled):
    """
    A wrapper to enforce types within a function or method given argument annotations.

    Each wrapped item is converted into a special `FunctionMethodEnforcer` class object that validates the passed parameters for the function or method when it is called. If a function or method that is passed does not have any annotations, it is not converted into a `FunctionMethodEnforcer` class as no validation is possible.

    If wrapping a class, all methods in the class that meet any of the following criteria will be wrapped individually:

    - Methods with `__call__`
    - Methods wrapped with `staticmethod` (if python >= 3.10)
    - Methods wrapped with `classmethod` (if python >= 3.10)

    Requires:

    - `clsFnMethod`:
        - What: The class, function or method that should have input types enforced
        - Type: function | method | class

    Example Use:
    ```
    >>> import type_enforced
    >>> @type_enforced.Enforcer
    ... def my_fn(a: int , b: [int, str] =2, c: int =3) -> None:
    ...     pass
    ...
    >>> my_fn(a=1, b=2, c=3)
    >>> my_fn(a=1, b='2', c=3)
    >>> my_fn(a='a', b=2, c=3)
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "/home/conmak/development/personal/type_enforced/type_enforced/enforcer.py", line 85, in __call__
        self.__check_type__(assigned_vars.get(key), value, key)
      File "/home/conmak/development/personal/type_enforced/type_enforced/enforcer.py", line 107, in __check_type__
        self.__exception__(
      File "/home/conmak/development/personal/type_enforced/type_enforced/enforcer.py", line 34, in __exception__
        raise Exception(f"({self.__fn__.__qualname__}): {message}")
    Exception: (my_fn): Type mismatch for typed variable `a`. Expected one of the following `[<class 'int'>]` but got `<class 'str'>` instead.
    ```
    """
    if not hasattr(clsFnMethod, "__enforcer_enabled__"):
        clsFnMethod.__enforcer_enabled__ = enabled
    if clsFnMethod.__enforcer_enabled__ == False:
        return clsFnMethod
    # Support eager annotations for python 3.14+
    if hasattr(clsFnMethod, "__annotate__"):
        if (
            not hasattr(clsFnMethod, "__annotations__")
            and getattr(clsFnMethod, "__annotate__") is not None
        ):
            clsFnMethod.__annotations__ == clsFnMethod.__annotate__(2)
    if isinstance(
        clsFnMethod, (staticmethod, classmethod, FunctionType, MethodType)
    ):
        # Only apply the enforcer if annotations are specified
        if getattr(clsFnMethod, "__annotations__", {}) == {}:
            return clsFnMethod
        elif isinstance(clsFnMethod, staticmethod):
            return staticmethod(FunctionMethodEnforcer(clsFnMethod.__func__))
        elif isinstance(clsFnMethod, classmethod):
            return classmethod(FunctionMethodEnforcer(clsFnMethod.__func__))
        elif isinstance(clsFnMethod, (FunctionType, MethodType)):
            return FunctionMethodEnforcer(clsFnMethod)
    elif hasattr(clsFnMethod, "__dict__"):
        for key, value in clsFnMethod.__dict__.items():
            # Skip the __annotate__ method if present in __dict__
            if key == "__annotate__":
                continue
            if hasattr(value, "__call__") or isinstance(
                value, (classmethod, staticmethod)
            ):
                setattr(clsFnMethod, key, Enforcer(value, enabled=enabled))
        return clsFnMethod
    else:
        raise Exception(
            "Enforcer can only be used on class methods, functions, or classes."
        )


Enforcer = Partial(Enforcer, enabled=True)
